// =====================================================================
// kstream.c
// =====================================================================

// Author:         (c) 2013 Daniele Cono D'Elia, Camil Demetrescu
// License:        See the end of this file for license information
// Created:        June 21, 2013

// Last changed:   $Date: 2013/06/20 13:01:07 $
// Changed by:     $Author: demetres $
// Revision:       $Revision: 1.00 $


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "kstream.h"


#ifndef DEBUG
#define DEBUG 1
#endif

#ifndef ADD_TO_TAIL
#define ADD_TO_TAIL 0
#endif


// alias
typedef kf_node tnode_t;

// structure definition
struct kstream {
    char*     name;
    tnode_t** roots;
    tnode_t*  top;
    tnode_t*  bottom;  
    uint32_t  stream_len;
    uint32_t  k;
    uint32_t  num_ids;
};


// used for building the kf from the ksf
typedef struct lnode_t lnode_t;
struct lnode_t {
    tnode_t* elem;
    lnode_t* next;    
};


// private function prototypes
static uint32_t _dump_aux(tnode_t* root, tnode_t* parent, FILE* out);
static void _populate_ids(tnode_t* root, lnode_t* ids[]);
static void _tree_delete(tnode_t* root);
static void _tree_join(tnode_t* node, tnode_t* ipf_root, uint32_t depth);

#if DEBUG
static void _tree_print(tnode_t* node, uint32_t depth);
#endif

static uint32_t _build_kf_stat_rec(kf_stat* stat, kf_node* node, 
                                   double* ratio_sum);


// ---------------------------------------------------------------------
// kstream_new
// ---------------------------------------------------------------------
kstream* kstream_new(char* name, uint32_t num_ids, uint32_t k) {

    long i;
    kstream* s = calloc(sizeof(kstream), 1);

    s->name    = name;
    s->num_ids = num_ids;
    s->k       = k;  
    s->roots   = malloc(num_ids*sizeof(tnode_t*));

    // create array of roots (singletons), one per distinct path id
    for (i = 0; i < num_ids; ++i) {
        s->roots[i] = calloc(sizeof(tnode_t), 1);
        s->roots[i]->id = i;
    }

    return s;      
}


// ---------------------------------------------------------------------
// kstream_start
// ---------------------------------------------------------------------
void kstream_start(kstream* s) {
    s->stream_len = 0;
    s->top = NULL;
}


// ---------------------------------------------------------------------
// kstream_emit
// ---------------------------------------------------------------------
void kstream_emit(kstream* s, uint32_t id) {  

    tnode_t* child;

    if (s->stream_len % (s->k-1) == 0) {
        s->bottom = s->top;
        s->top = s->roots[id];
    }

    else {

        for (child = s->top->first_child; 
             child != NULL; 
             child = child->next_sibling) if (child->id == id) break;

        if (child == NULL) {
            child = calloc(sizeof(tnode_t), 1);
            child->id = id;
            child->next_sibling = s->top->first_child;
            s->top->first_child = child;
        }
        
        s->top=child;    
    }

    if (s->bottom != NULL) {

        for (child = s->bottom->first_child; 
             child != NULL; 
             child = child->next_sibling) if (child->id == id) break;

        if (child == NULL) {
            child = calloc(sizeof(tnode_t), 1);
            child->id = id;
            child->next_sibling = s->bottom->first_child;
            s->bottom->first_child = child;
        }

        child->count++;
        s->bottom=child;
    }
    
    else {
    
        s->top->count++;
                
    }    

    s->stream_len++;
}


// ---------------------------------------------------------------------
// kstream_dump
// ---------------------------------------------------------------------
void kstream_dump(kstream* s, char* log_file) {

    FILE* out = fopen(log_file, "w");
    
    if (out == NULL) {
        fprintf(stderr, "[kstream] unable to create logfile %s\n", log_file);
        return;
    }
    
    // TODO - for compatibility with ProcessLog.java (k-1)
    fprintf(out, "c kblpp %u\n", s->k-1); 
    
    long i;
    uint32_t num_nodes = 0;  
    
    // TODO - for compatibility with ProcessLog.java (cleans counters)
    fprintf(out, "v %lx %lx 0 %u %u %u 0\n", (unsigned long)s->roots[0], 
            (unsigned long)NULL, s->roots[0]->id, s->roots[0]->count, 1);
    ++num_nodes;
    tnode_t* tmp;
    for (tmp=s->roots[0]->first_child; tmp!=NULL; tmp=tmp->next_sibling)
        num_nodes += _dump_aux(tmp, s->roots[0], out);
    
    for (i = 1; i < s->num_ids; ++i)
        num_nodes += _dump_aux(s->roots[i], NULL, out);
    fprintf(out, "p %u\n", num_nodes);
    
    fclose(out);  
}    


// ---------------------------------------------------------------------
// kstream_delete
// ---------------------------------------------------------------------
void kstream_delete(kstream* s) {
    long i;
    for (i = 0; i < s->num_ids ; ++i) _tree_delete(s->roots[i]);
    free(s->roots);
    free(s);    
}


// ---------------------------------------------------------------------
// kf_new
// ---------------------------------------------------------------------
kf_node** kf_new(kstream* s) {

    tnode_t** kf  = calloc(sizeof(tnode_t*)*s->num_ids, 1);
    lnode_t** ids = calloc(sizeof(lnode_t*)*s->num_ids, 1);
    
    // temp variables
    long i;
    tnode_t* child;
    lnode_t* l;

    #if DEBUG
    printf("Printing ksf after counter cleaning...\n");
    for (i = 0; i < s->num_ids; ++i) _tree_print(s->roots[i], 1);
    printf("Done!\n");
    #endif
    
    // populate ids
    for (i = 0; i < s->num_ids; ++i)    
        _populate_ids(s->roots[i], ids);

    // build kf from ksf
    tnode_t* dummy_root = calloc(sizeof(tnode_t), 1);
    #if ADD_TO_TAIL
    for (i = 0; i < s->num_ids; ++i)
        for (l = ids[i]; l != NULL; l = l->next)
            _tree_join(l->elem, dummy_root, s->k);
    #else
    for (i = s->num_ids - 1 ; i >= 0 ; --i)
        for (l = ids[i]; l != NULL; l = l->next)
            _tree_join(l->elem, dummy_root, s->k);
    #endif        

    // prepare array of pointers to return
    for (child = dummy_root->first_child; 
         child != NULL; 
         child = child->next_sibling) kf[child->id] = child;
    
    // deallocate temporary list
    for (i = 0; i < s->num_ids; ++i) {
        l = ids[i];    
        while (l != NULL) {
            lnode_t* dead = l;
            l = l->next;
            free(dead);
        }
    }

    free(ids);
    free(dummy_root);
    
    #if DEBUG
    printf("\nPrinting kf...\n");
    for (i = 0; i < s->num_ids; ++i) _tree_print(kf[i], 1);
    printf("Done!\n");
    #endif
    
    return kf;  
}


// ---------------------------------------------------------------------
// kf_delete
// ---------------------------------------------------------------------
void kf_delete(tnode_t** kf) {
    tnode_t* p = kf[0];
    while (p != NULL) {        
        tnode_t* dead = p;
        p = p->next_sibling;
        _tree_delete(dead);
    }
    free(kf);
}


// ---------------------------------------------------------------------
// kf_get_stat
// ---------------------------------------------------------------------
void kf_get_stat(kf_node** kf, kf_stat* stat) {
    
    kf_node* node;
    double ratio_sum = 0;
    
    // reset stats
    memset(stat, 0, sizeof(kf_stat));
    
    // scan trees in the forest
    for (node = kf[0]; node != NULL; node = node->next_sibling) {

        // count trees in the forest
        stat->num_trees++;
        
        // compute tree statistics
        uint32_t num_levels = _build_kf_stat_rec(stat, node, &ratio_sum);
        
        // update forest depth
        if (num_levels > stat->num_levels)
            stat->num_levels = num_levels;
    }
    
    // compute counter skewness
    stat->counter_skewness = stat->num_int_nodes ? 
                                ratio_sum / stat->num_int_nodes : 0;
}


// ---------------------------------------------------------------------
// kf_print_stat
// ---------------------------------------------------------------------
void kf_print_stat(kf_stat* stat){
    printf("Number of trees: %u\n", (unsigned)stat->num_trees);
    printf("Number of nodes: %u\n", (unsigned)stat->num_nodes);
    printf("Number of internal nodes: %u\n", (unsigned)stat->num_int_nodes);
    printf("Number of levels: %u\n", (unsigned)stat->num_levels);
    printf("Average counter skewness: %.3f\n", stat->counter_skewness);
}


// ---------------------------------------------------------------------
// _dump_aux
// ---------------------------------------------------------------------
static uint32_t _dump_aux(tnode_t* root, tnode_t* parent, FILE* out) {

    // flag: 0 NORMAL, 1 BEGIN, 2 EXIT
    uint32_t flag = 0;

    fprintf(out, "v %lx %lx 0 %u %u %u 0\n", (unsigned long)root, 
            (unsigned long)parent, root->id, root->count, flag);
    
    tnode_t* child;
    uint32_t ret = 0;
    for (child = root->first_child; 
         child != NULL; 
         child = child->next_sibling)
        ret += _dump_aux(child, root, out);

    return ret+1;
}    


// ---------------------------------------------------------------------
// _populate_ids
// ---------------------------------------------------------------------
static void _populate_ids(tnode_t* root, lnode_t* ids[]) {
    
    lnode_t* l;  
    
    l = malloc(sizeof(lnode_t));
    l->elem = root;
    l->next = ids[root->id];
    ids[root->id] = l;
    
    tnode_t* child;
    for (child = root->first_child; 
         child != NULL; 
         child = child->next_sibling) {        
        
        _populate_ids(child, ids);
    }        
}


// ---------------------------------------------------------------------
// _tree_delete
// ---------------------------------------------------------------------
// recursively delete tree
static void _tree_delete(tnode_t* root) { 

    tnode_t* child = root->first_child;

    while (child != NULL) {
        tnode_t* dead = child;
        child = child->next_sibling;
        _tree_delete(dead);
    }  

    free(root);
}    


// ---------------------------------------------------------------------
// _tree_join
// ---------------------------------------------------------------------
static void _tree_join(tnode_t* node, tnode_t* ipf_root, uint32_t depth) {

    tnode_t* ipf_child;
    tnode_t* tmp;

    for (ipf_child = ipf_root->first_child; 
         ipf_child != NULL; 
         ipf_child = ipf_child->next_sibling)
        if (ipf_child->id == node->id) break;
    
    // if node does not exist, create and insert it correctly
    if (ipf_child == NULL) {

        // create child
        ipf_child = calloc(sizeof(tnode_t), 1);
        ipf_child->id = node->id;    
        
        #if ADD_TO_TAIL
        // add to tail
        if (ipf_root->first_child == NULL)
            ipf_root->first_child = ipf_child;

        else {
            for (tmp = ipf_root->first_child; 
                 tmp->next_sibling != NULL; 
                 tmp = tmp->next_sibling);
            tmp->next_sibling = ipf_child;
        }
        #else
        // add to front
        ipf_child->next_sibling = ipf_root->first_child;
        ipf_root->first_child = ipf_child;    
        #endif
    }
    
    ipf_child->count += node->count; // branchless :)    

    // process subtree having node as root
    if (depth > 1) {
        for (tmp = node->first_child; 
             tmp != NULL; 
             tmp = tmp->next_sibling)
            _tree_join(tmp, ipf_child, depth-1);    
    }      
}


// ---------------------------------------------------------------------
// _tree_print
// ---------------------------------------------------------------------
#if DEBUG
static void _tree_print(tnode_t* root, uint32_t depth) {
    long i;
    tnode_t* child;
    
    // print root
    for (i = 0; i<depth; ++i) printf("--");
    printf("> %d (%d)\n", root->id, root->count);
    
    for (child = root->first_child; 
         child != NULL; 
         child = child->next_sibling)
        _tree_print(child, depth+1);
}    
#endif


// ---------------------------------------------------------------------
// _build_kf_stat_rec
// ---------------------------------------------------------------------
static uint32_t _build_kf_stat_rec(kf_stat* stat, kf_node* node, 
                                   double* ratio_sum) {

    uint32_t max_num_levels = 0;
    double max_ratio = 0;
    kf_node* child;

    if (node == NULL) return 0;

    // update node counters
    stat->num_nodes++;
    if (node->first_child != NULL) stat->num_int_nodes++;

    // scan node children
    for (child = node->first_child; 
         child != NULL; 
         child = child->next_sibling) {

        // compute ratio between node counter and child counter
        double ratio = (double)child->count/node->count;

        // keep track of max ratio over all children
        if (ratio > max_ratio) max_ratio = ratio;

        // recurse on child
        uint32_t num_levels = _build_kf_stat_rec(stat, child, ratio_sum);

        // update tree levels
        if (num_levels > max_num_levels) max_num_levels = num_levels;
    }

    *ratio_sum += max_ratio;
    
    return 1 + max_num_levels;
}


// Copyright (C) 2013 Danile Cono D'Elia, Camil Demetrescu

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
// USA

